# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import os

import lightning.pytorch

from lightning.pytorch import Trainer

from .callbacks import PredictionsWriter
from .device import DeviceManager

logger = logging.getLogger(__name__)


def run(
    model: lightning.pytorch.LightningModule,
    datamodule: lightning.pytorch.LightningDataModule,
    device_manager: DeviceManager,
    output_folder: str,
):
    """Runs inference on input data, outputs csv files with predictions.

    Parameters
    ---------
    model : :py:class:`torch.nn.Module`
        Neural network model (e.g. pasa).

    datamodule
        The lightning datamodule to use for training **and** validation

    device_manager
        An internal device representation, to be used for training and
        validation.  This representation can be converted into a pytorch device
        or a torch lightning accelerator setup.

    output_folder : str
        Directory in which the results will be saved.

    grad_cams : bool
        If we export grad cams for every prediction (must be used along
        a batch size of 1 with the DensenetRS model).

    Returns
    -------

    all_predictions : list
        All the predictions associated with filename and ground truth.
    """

    logger.info(f"Output folder: {output_folder}")
    os.makedirs(output_folder, exist_ok=True)

    logfile_fields = ("filename", "likelihood", "ground_truth")

    accelerator, devices = device_manager.lightning_accelerator()
    trainer = Trainer(
        accelerator=accelerator,
        devices=devices,
        callbacks=[
            PredictionsWriter(
                output_dir=output_folder,
                logfile_fields=logfile_fields,
                write_interval="epoch",
            ),
        ],
    )

    all_predictions = trainer.predict(model, datamodule)

    return all_predictions
