# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Tools for interacting with the running computer or GPU."""

import logging
import multiprocessing
import multiprocessing.synchronize
import os
import queue
import shutil
import subprocess
import time
import typing

import numpy
import psutil

logger = logging.getLogger(__name__)

_nvidia_smi = shutil.which("nvidia-smi")
"""Location of the nvidia-smi program, if one exists."""


GB = float(2**30)
"""The number of bytes in a gigabyte."""


def run_nvidia_smi(
    query: typing.Sequence[str],
) -> dict[str, str | float] | None:
    """Returns GPU information from query.

    For a comprehensive list of options and help, execute ``nvidia-smi
    --help-query-gpu`` on a host with a GPU


    Parameters
    ----------

    query
        A list of query strings as defined by ``nvidia-smi --help-query-gpu``


    Returns
    -------

    data
        A dictionary containing the queried parameters (``rename`` versions).
        If ``nvidia-smi`` is not available, returns ``None``.  Percentage
        information is left alone, memory information is transformed to
        gigabytes (floating-point).
    """
    if _nvidia_smi is None:
        return None

    # Gets GPU information, based on a GPU device if that is set. Returns
    # ordered results.
    query_str = (
        f"{_nvidia_smi} --query-gpu={','.join(query)} --format=csv,noheader"
    )
    visible_devices = os.environ.get("CUDA_VISIBLE_DEVICES")
    if visible_devices:
        query_str += f" --id={visible_devices}"
    values = subprocess.getoutput(query_str)

    retval: dict[str, str | float] = {}
    for i, k in enumerate([k.strip() for k in values.split(",")]):
        retval[query[i]] = k
        if k.endswith("%"):
            retval[query[i]] = float(k[:-1].strip())
        elif k.endswith("MiB"):
            retval[query[i]] = float(k[:-3].strip()) / 1024
    return retval


def gpu_constants() -> dict[str, str | int | float] | None:
    """Returns GPU (static) information using nvidia-smi.

    See :py:func:`run_nvidia_smi` for operational details.

    Returns
    -------

    data : :py:class:`tuple`, None
        If ``nvidia-smi`` is not available, returns ``None``, otherwise, we
        return a dictionary containing the following ``nvidia-smi`` query
        information, in this order:

        * ``gpu_name``, as ``gpu_name`` (:py:class:`str`)
        * ``driver_version``, as ``gpu_driver_version`` (:py:class:`str`)
        * ``memory.total``, as ``gpu_memory_total`` (transformed to gigabytes,
          :py:class:`float`)
    """
    retval = run_nvidia_smi(("gpu_name", "driver_version", "memory.total"))
    if retval is None:
        return retval

    # else, just update with more generic names
    retval["driver-version/gpu"] = retval.pop("driver_version")
    retval["total-memory-GB/gpu"] = retval.pop("memory.total")
    return retval


def gpu_log() -> dict[str, float] | None:
    """Returns GPU information about current non-static status using nvidia-
    smi.

    See :py:func:`run_nvidia_smi` for operational details.

    Returns
    -------

    data
        If ``nvidia-smi`` is not available, returns ``None``, otherwise, we
        return a dictionary containing the following ``nvidia-smi`` query
        information, in this order:

        * ``memory.used``, as ``gpu_memory_used`` (transformed to gigabytes,
          :py:class:`float`)
        * ``memory.free``, as ``gpu_memory_free`` (transformed to gigabytes,
          :py:class:`float`)
        * ``100*memory.used/memory.total``, as ``gpu_memory_percent``,
          (:py:class:`float`, in percent)
        * ``utilization.gpu``, as ``gpu_percent``,
          (:py:class:`float`, in percent)
    """

    result = run_nvidia_smi(
        ("memory.total", "memory.used", "memory.free", "utilization.gpu")
    )

    if result is None:
        return result

    return {
        "memory-used-GB/gpu": float(result["memory.used"]),
        "memory-free-GB/gpu": float(result["memory.free"]),
        "memory-percent/gpu": 100
        * float(result["memory.used"])
        / float(result["memory.total"]),
        "percent-usage/gpu": float(result["utilization.gpu"]),
    }


def cpu_constants() -> dict[str, int | float]:
    """Returns static CPU information about the current system.

    Returns
    -------

    data
        An ordered dictionary (organized as 2-tuples) containing these entries:

        0. ``cpu_memory_total`` (:py:class:`float`): total memory available,
           in gigabytes
        1. ``cpu_count`` (:py:class:`int`): number of logical CPUs available
    """
    return {
        "memory-total-GB/cpu": psutil.virtual_memory().total / GB,
        "count/cpu": psutil.cpu_count(logical=True),
    }


class CPULogger:
    """Logs CPU information using :py:mod:`psutil`

    Parameters
    ----------

    pid
        Process identifier of the main process (parent process) to observe
    """

    def __init__(self, pid: int | None = None):
        this = psutil.Process(pid=pid)
        self.cluster = [this] + this.children(recursive=True)
        # touch cpu_percent() at least once for all processes in the cluster
        [k.cpu_percent(interval=None) for k in self.cluster]

    def log(self) -> dict[str, int | float]:
        """Returns current process cluster iformation.

        Returns
        -------

        data
            An ordered dictionary containing these entries:

            0. ``cpu_memory_used`` (:py:class:`float`): total memory used from
               the system, in gigabytes
            1. ``cpu_rss`` (:py:class:`float`):  RAM currently used by
               process and children, in gigabytes
            2. ``cpu_vms`` (:py:class:`float`):  total memory (RAM + swap) currently
               used by process and children, in gigabytes
            3. ``cpu_percent`` (:py:class:`float`): percentage of the total CPU
               used by this process and children (recursively) since last call
               (first time called should be ignored).  This number depends on the
               number of CPUs in the system and can be greater than 100%
            4. ``cpu_processes`` (:py:class:`int`): total number of processes
               including self and children (recursively)
            5. ``cpu_open_files`` (:py:class:`int`): total number of open files by
               self and children
        """
        # check all cluster components and update process list
        # done so we can keep the cpu_percent() initialization
        stored_children = set(self.cluster[1:])
        current_children = set(self.cluster[0].children(recursive=True))
        keep_children = stored_children - current_children
        new_children = current_children - stored_children
        gone = set()
        for k in new_children:
            try:
                k.cpu_percent(interval=None)
            except (psutil.ZombieProcess, psutil.NoSuchProcess):
                # child process is gone meanwhile
                # update the intermediate list for this time
                gone.add(k)
        new_children = new_children - gone
        self.cluster = (
            self.cluster[:1] + list(keep_children) + list(new_children)
        )

        memory_info = []
        cpu_percent = []
        open_files = []
        gone = set()

        for k in self.cluster:
            try:
                memory_info.append(k.memory_info())
                cpu_percent.append(k.cpu_percent(interval=None))
                open_files.append(len(k.open_files()))
            except (psutil.ZombieProcess, psutil.NoSuchProcess):
                # child process is gone meanwhile, just ignore it
                # it is too late to update any intermediate list
                # at this point, but ensures to update counts later on
                gone.add(k)
        return {
            "memory-used-GB/cpu": psutil.virtual_memory().used / GB,
            "rss-GB/cpu": sum([k.rss for k in memory_info]) / GB,
            "vms-GB/cpu": sum([k.vms for k in memory_info]) / GB,
            "percent-usage/cpu": sum(cpu_percent),
            "num-processes/cpu": len(self.cluster) - len(gone),
            "num-open-files/cpu": sum(open_files),
        }


class _InformationGatherer:
    """A container to store monitoring information.

    Parameters
    ----------

    has_gpu
        A flag indicating if we have a GPU installed on the platform or not

    main_pid
        The main process identifier to monitor

    logger
        A logger to be used for logging messages
    """

    def __init__(
        self, has_gpu: bool, main_pid: int | None, logger: logging.Logger
    ):
        self.logger: logging.Logger = logger
        self.cpu_logger = CPULogger(main_pid)
        keys: list[str] = list(self.cpu_logger.log().keys())
        self.has_gpu: bool = has_gpu
        if self.has_gpu:
            example = gpu_log()
            if example is not None:
                keys += list(example.keys())
        self.data: dict[str, list[int | float]] = {k: [] for k in keys}

    def acc(self) -> None:
        """Accumulates another measurement."""
        for k, v in self.cpu_logger.log().items():
            self.data[k].append(v)
        if self.has_gpu:
            sample = gpu_log()
            if sample is not None:
                for k, v in sample.items():
                    self.data[k].append(v)

    def clear(self) -> None:
        """Clears accumulated data."""
        for k in self.data.keys():
            self.data[k] = []

    def summary(self) -> dict[str, list[int | float]]:
        """Returns the current data."""
        if len(next(iter(self.data.values()))) == 0:
            self.logger.error("CPU/GPU logger was not able to collect any data")
        return self.data


def _monitor_worker(
    interval: int | float,
    has_gpu: bool,
    main_pid: int,
    stop: multiprocessing.synchronize.Event,
    summary_event: multiprocessing.synchronize.Event,
    queue: queue.Queue,
    logging_level: int,
):
    """A monitoring worker that measures resources and returns lists.

    Parameters
    ==========

    interval
        Number of seconds to wait between each measurement (maybe a floating
        point number as accepted by :py:func:`time.sleep`)

    has_gpu
        A flag indicating if we have a GPU installed on the platform or not

    main_pid
        The main process identifier to monitor

    stop
        Event that indicates if we should continue running or stop

    summary_event
        Event that indicates if we should produce a summary

    queue
        A queue, to send monitoring information back to the spawner

    logging_level
        The logging level to use for logging from launched processes
    """
    logger = multiprocessing.log_to_stderr(level=logging_level)
    ra = _InformationGatherer(has_gpu, main_pid, logger)

    while not stop.is_set():
        try:
            ra.acc()  # guarantees at least an entry will be available

            if summary_event.is_set():
                queue.put(ra.summary().copy())
                ra.clear()
                summary_event.clear()

            time.sleep(interval)
        except Exception:
            logger.exception(
                "Iterative CPU/GPU logging did not work properly."
                " Exception follows.  Retrying..."
            )
            time.sleep(0.5)  # wait half a second, and try again!


class ResourceMonitor:
    """An external, non-blocking CPU/GPU resource monitor.

    Parameters
    ----------

    interval
        Number of seconds to wait between each measurement (maybe a floating
        point number as accepted by :py:func:`time.sleep`)

    has_gpu
        A flag indicating if we have a GPU installed on the platform or not

    main_pid
        The main process identifier to monitor

    logging_level
        The logging level to use for logging from launched processes
    """

    def __init__(
        self,
        interval: int | float,
        has_gpu: bool,
        main_pid: int,
        logging_level: int,
    ):
        self.interval = interval
        self.has_gpu = has_gpu
        self.main_pid = main_pid
        self.stop_event = multiprocessing.Event()
        self.summary_event = multiprocessing.Event()
        self.q: multiprocessing.Queue[
            dict[str, list[int | float]]
        ] = multiprocessing.Queue()
        self.logging_level = logging_level

        self.monitor = multiprocessing.Process(
            target=_monitor_worker,
            name="ResourceMonitorProcess",
            args=(
                self.interval,
                self.has_gpu,
                self.main_pid,
                self.stop_event,
                self.summary_event,
                self.q,
                self.logging_level,
            ),
        )

        self.data: dict[str, int | float] | None = None

    def __enter__(self) -> None:
        """Starts the monitoring process."""
        self.monitor.start()

    def checkpoint(self) -> None:
        """Forces the monitoring process to yield data and clear the internal
        accumlator."""
        self.summary_event.set()
        try:
            data: dict[str, list[int | float]] = self.q.get(
                timeout=2 * self.interval
            )
        except queue.Empty:
            logger.warning(
                f"CPU/GPU resource monitor did not provide anything when "
                f"joined (even after a {2*self.interval}-second timeout - "
                f"this is normally due to exceptions on the monitoring process. "
                f"Check above for other exceptions."
            )
            self.data = None
        else:
            # summarize the returned data by creating averages
            self.data = {}
            for k, values in data.items():
                if values:
                    if k in ("cpu_processes", "cpu_open_files"):
                        self.data[k] = numpy.max(values)
                    else:
                        self.data[k] = float(numpy.mean(values))
                else:
                    self.data[k] = 0.0

    def __exit__(self, *_) -> None:
        """Stops the monitoring process and returns the summary of
        observations."""

        self.stop_event.set()
        self.monitor.join()
        if self.monitor.exitcode != 0:
            logger.error(
                f"CPU/GPU resource monitor process exited with code "
                f"{self.monitor.exitcode}.  Check logs for errors!"
            )
